/* EeasyShell - conceptual UNIX micro-shell with job management.
 *
 * Copyright 2010, Grisha Sitkarev
 * <sitkarev@unixkomi.ru>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <limits.h>

#include "macros.h"

#define SPACES		"\t "
#define PROMPT		"$"

#define NRARGS_MAX	15

enum {
	PGRP_STOPPED = 1,
	PGRP_RUNNING = 2
};

struct pgrp {
	struct pgrp     *next;
	pid_t           pgid;
	char            cmd[NAME_MAX+1];
	int             status;
	int		flags;
};

static char *program_name;

static int terminate;
static int interrupted;

static struct pgrp *background_list;
static struct pgrp *completed_list;

static int recent_status;

void
sighandler(int signo)
{
	switch (signo) {
	case SIGINT:
	case SIGQUIT:
		++interrupted;
		break;
	case SIGTERM:
		++terminate;
		break;
	}
}

/* Prints shell prompt. */
void
prompt()
{
	printf("%s ", PROMPT);
	fflush(stdout);
}

/* Reads command line and parses it into argument list args. */
void
read_cmd(char *buf, int *nargs, char *args[])
{
	int n, cnt;
	char *s, *end;

	*nargs = -1;

	s = fgets(buf, LINE_MAX, stdin);
	if (s == NULL)
		return;

	n = strlen(s);
	while ((n > 0) && (s[n-1] == '\n'))
		s[--n] = '\0';

	end = s + n;
	cnt = 0;

	while (s != end && cnt < NRARGS_MAX) {
		s += strspn(s, SPACES);
		if (s == end)
			break;
		args[cnt] = s;
		s += strcspn(s, SPACES);
		*s = '\0';
		if (s != end)
			++s;
		++cnt;
	}

	args[cnt] = NULL;
	*nargs = cnt;
}

/* Ensures that stdin and stdout is a terminal device. */
void
ensure_tty()
{
	if (!isatty(STDIN_FILENO) || !isatty(STDOUT_FILENO)) {
		ERROR("not a tty");
		exit(1);
	}
}

void
set_program_name(int argc, char *argv[])
{
	char *s;

	if (argc <= 0) {
		program_name = "???";
		return;
	}

	s = strrchr(argv[0], '/');

	if (s != NULL)
		program_name = ++s;
	else
		program_name = argv[0];
}

/* Initializes signals disposition. */
void
set_signals()
{
	int i, res;
	struct sigaction act;

	struct sigtab {
		int signo;
		void (*handler)(int);
	} sigs[] = {
		{ SIGTTOU, SIG_IGN },
		{ SIGINT, sighandler },
		{ SIGQUIT, sighandler },
		{ SIGTERM, sighandler }
	};

	memset(&act, 0, sizeof(act));

	for (i = 0; i < ARRAY_SIZE(sigs); i++) {
		sigemptyset(&act.sa_mask);
		act.sa_flags = 0;
		act.sa_handler = sigs[i].handler;
		res = sigaction(sigs[i].signo, &act, NULL);
		if (res == -1) {
			SYS_ERROR("sigaction() signo %d", sigs[i].signo);
		}
	}
}

/* fork->exec command. Returns pgrp structure on success. */
struct pgrp *
run_cmd(int nargs, char *args[])
{
	struct pgrp *pg;
	pid_t pid;
	int res;

	if (nargs <= 0)
		return NULL;

	if (access(args[0], R_OK | X_OK) != 0) {
		ERR_MSG("%s", args[0]);
		return NULL;
	}

	pid = fork();
	switch (pid) {
	case 0:
		/* Make this child a group leader. */
		res = setpgid(0, 0);
		/* Restore disabled SIGTTOU. */
		signal(SIGTTOU, SIG_DFL);
		execv(args[0], args);
		ERR_MSG("execl: %s", args[0]);
		_exit(1);
		break;
	case -1:
		ERR_MSG("fork: %s", args[0]);
		return NULL;
	default:
		/* Parent also assists in making child a group leader.
		   This prevents race condition. */
		res = setpgid(pid, pid);
		if (res == -1)
			SYS_ERROR("setpgid: %s", args[0]);
		pg = malloc(sizeof(*pg));
		if (pg == NULL) {
			ERROR("can't alloc pg");
			exit(2);
		}
		memset(pg, 0, sizeof(*pg));
		strncpy(pg->cmd, args[0], sizeof(pg->cmd));
		pg->cmd[sizeof(pg->cmd)-1] = '\0';
		pg->pgid = pid;
		pg->flags = PGRP_RUNNING;
		pg->status = 0;
		break;
	}

	return pg;
}

/* Brings process group to foreground. */
void
bring_to_fg(pid_t cpgid)
{
	int res;

	/* Bring this group to foreground. */
	res = tcsetpgrp(STDIN_FILENO, cpgid);
	if (res == -1)
		SYS_ERROR("tcsetpgrp");
	/* Continue processes if stopped. */
	res = killpg(cpgid, SIGCONT);
	if (res == -1)
		SYS_ERROR("killpg");
}

/* Waits for foreground process group. */
void
wait_for_pg(struct pgrp *pg)
{
	int status;
	pid_t pid;

	while ((pid = waitpid(-pg->pgid, &status, WUNTRACED)) > 0) {

		/* Is this process a group leader? */
		if (pid != pg->pgid)
			continue;

		if (WIFEXITED(status)) {
			recent_status = WEXITSTATUS(status);
			free(pg);
		} else if (WIFSIGNALED(status)) {
			recent_status = WTERMSIG(status)+128;
			free(pg);
		} else if (WIFSTOPPED(status)) {
			/* Add this process group to the stopped list. */
			printf("\n");
			printf("[%d]    STOPPED    %.60s\n", pid, pg->cmd);
			pg->flags = PGRP_STOPPED;
			pg->next = background_list;
			background_list = pg;
		}

		break;
	}
}

/* Updates status of all background and stopped commands (groups). */
void
update_bg_pgrp_status()
{
	pid_t pid;
	int status;

	while ((pid = waitpid(-1, &status, WNOHANG|WUNTRACED)) > 0) {
		struct pgrp *pg, *prev;
		int completed = 0;

		/* Proceed with status change only if group leader changed. */
		prev = NULL;
		for (pg = background_list; pg != NULL; pg = pg->next) {
			if (pg->pgid == pid)
				break;
			prev = pg;
		}

		if (pg == NULL)
			continue;

		if (pg->flags == PGRP_STOPPED) {
			ERROR("stopped pgid=%d changed status?", pid);
			continue;
		}

		if (WIFEXITED(status)) {
			pg->status = WEXITSTATUS(status);
			completed = 1;
		} else if (WIFSIGNALED(status)) {
			pg->status = WTERMSIG(status)+128;
			completed = 1;
		} else if (WIFSTOPPED(status)) {
			pg->status = WSTOPSIG(status);
			pg->flags = PGRP_STOPPED;
		}

		/* Remove this pgrp from list. */
		if (completed) {
			if (prev != NULL)
				prev->next = pg->next;
			else
				background_list = pg->next;
			pg->next = completed_list;
			completed_list = pg;
		}
	}
}

/* Executes internal shell command (if found). */
int
try_internal_cmd(int nargs, char *args[], struct pgrp **fg)
{
	struct pgrp *pg, *prev;
	pid_t pgid;
	int res;
	
	if (nargs == 0)
		return 0;

	*fg = NULL;

	if (strcmp(args[0], "jobs") == 0) {
		char *s;
		/* Update jobs status before listing it. We can
		   miss job with SIGTTIN/SIGTTOU/SIGSTOP
		   after change from stopped to running. */
		update_bg_pgrp_status();

		for (pg = background_list; pg != NULL; pg = pg->next) {
			switch (pg->flags) {
			case PGRP_STOPPED:
				s = "STOPPED";
				break;
			case PGRP_RUNNING:
				s = "RUNNING";
				break;
			default:
				s = "???";
				break;
			}
			printf("[%d]    %s    %.60s\n", pg->pgid, s, pg->cmd);
		}
		return 0;
	} else if (strcmp(args[0], "fg") == 0) {
		if (nargs < 2) {
			printf("fg: missing required argument <pgid>\n");
			return 0;
		}
		pgid = atoi(args[1]);
		prev = NULL;
		for (pg = background_list; pg != NULL; pg = pg->next) {
			if (pg->pgid == pgid) {
				if (pg->flags == PGRP_STOPPED) {
					res = killpg(pg->pgid, SIGCONT);
					if (res == -1)
						SYS_ERROR("killpg");
					pg->flags = PGRP_RUNNING;
				}
				if (prev != NULL)
					prev->next = pg->next;
				else
					background_list = pg->next;
				*fg = pg;
				return 0;
			}
			prev = pg;
		}
		printf("fg: no such pgid %d\n", pgid);
		return 0;
	} else if (strcmp(args[0], "bg") == 0) {
		if (nargs < 2) {
			printf("bg: missing required argument <pgid>\n");
			return 0;
		}
		pgid = atoi(args[1]);
		for (pg = background_list; pg != NULL; pg = pg->next) {
			if (pg->pgid == pgid) {
				if (pg->flags != PGRP_STOPPED) {
					printf("bg: pgid %d is bg\n", pgid);
					return 0;
				}
				res = killpg(pg->pgid, SIGCONT);
				if (res == -1)
					SYS_ERROR("killpg");
				pg->flags = PGRP_RUNNING;
				return 0;
			}
		}
		printf("bg: no such pgid %d\n", pgid);
		return 0;
	}

	return -1;
}

void
display_completed()
{
	struct pgrp *pg, *next;

	for (pg = completed_list; pg != NULL; pg = next) {
		next = pg->next;
		printf("[%d]    COMPLETED   %.60s\n", pg->pgid, pg->cmd);
		free(pg);
	}
	completed_list = NULL;
}

void
shell_to_fg()
{
	bring_to_fg(getpgid(0));
}

int
main(int argc, char *argv[])
{
	int res, nargs;
	char *args[NRARGS_MAX+1];
	char buf[LINE_MAX];
	struct pgrp *pg;

	set_program_name(argc, argv);

	ensure_tty();

	set_signals();

	while (!terminate) {

		update_bg_pgrp_status();

		display_completed();

		prompt();

		read_cmd(buf, &nargs, args);

		if (nargs < 0) {
			if (interrupted) {
				interrupted = 0;
				printf("\n");
				continue;
			} else {
				if (background_list != NULL) {
					printf("jobs running!\n");
					continue;
				} else
			 		break;
			}
		}

		if (nargs == 0)
			continue;

		res = try_internal_cmd(nargs, args, &pg);

		if (res != 0)
			pg = run_cmd(nargs, args);

		if (pg == NULL)
			continue;

		/* Bring command process group to foreground. */
		bring_to_fg(pg->pgid);

		/* Wait for foreground process group. */
		wait_for_pg(pg);

		/* Bring shell to foreground. */
		shell_to_fg();
        }

	return 0;
}
